from typing import Any, Dict, Optional, Tuple, Union

import gym
import numpy as np

from popgym.core.deck import RANKS, Deck
from popgym.core.env import POPGymEnv


def value_fn(hand):
    if hand[-1] > hand[-2]:
        return 1
    elif hand[-1] == hand[-2]:
        return 0
    else:
        return -1


class HigherLower(POPGymEnv):
    """A game of higher/lower. Given a deck of cards, the agent predicts whether the
    next card drawn from the deck is higher or lower than the last card drawn from
    the deck. A push results in zero reward, while a correct/incorrect guess result
    in 1/deck_size and -1/deck_size reward. The agent can learn to count cards to
    infer which cards are left in the deck, improving accuracy.

    Args:
        num_decks: The number of individual decks combined into a single deck.

    Returns:
        A gym environment
    """

    def __init__(self, num_decks=1):
        self.num_decks = num_decks
        self.deck = Deck(num_decks)
        self.deck.add_players("player")
        self.action_space = gym.spaces.Discrete(2)
        self.state = np.zeros(
            (
                len(
                    RANKS,
                )
            ),
            dtype=np.uint8,
        )
        self.observation_space = self.deck.get_obs_space(["ranks"])
        self.state_space = gym.spaces.Box(0, 1, self.state.shape)
        self.value_map = dict(zip(self.deck.ranks, range(len(self.deck.ranks))))
        self.deck_size = len(self.deck)

    def get_state(self):
        return (self.state.copy() / 4 / self.num_decks).astype(np.float32)

    def step(self, action):
        guess_higher = action == 0
        done = len(self.deck) <= 1

        self.deck.deal("player", 1)
        assert self.deck.hand_size("player") == 2
        curr_value, next_value = self.deck.show("player", ["ranks_idx"]).reshape(-1)
        self.state[next_value] += 1

        rew_scale = 1 / self.deck_size
        if next_value == curr_value:
            reward = 0
        elif (next_value > curr_value) == guess_higher:
            reward = rew_scale
        else:
            reward = -rew_scale

        self.deck.discard("player", 0)
        obs = self.deck.show("player", ["ranks_idx"]).item()

        return obs, reward, done, {}

    def render(self, mode="ascii"):
        return self.deck.visualize_idx(self.deck["player"])

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, Dict[str, Any]]]:
        super().reset(seed=seed)
        self.deck.reset(rng=self.np_random)
        self.deck.deal("player", 1)
        self.state = np.zeros(
            (
                len(
                    RANKS,
                )
            ),
            dtype=np.uint8,
        )
        obs = self.deck.show("player", ["ranks_idx"]).item()
        self.state[obs] = 1
        viz = np.concatenate(self.deck.show("player", ["suits", "ranks"]))
        if return_info:
            return obs, {"card": viz}

        return obs


class HigherLowerEasy(HigherLower):
    def __init__(self):
        super().__init__(num_decks=1)


class HigherLowerMedium(HigherLower):
    def __init__(self):
        super().__init__(num_decks=2)


class HigherLowerHard(HigherLower):
    def __init__(self):
        super().__init__(num_decks=3)



from typing import Any, Dict, Optional, Tuple, Union

import gym
import numpy as np

from popgym.core.deck import Deck
from popgym.core.env import POPGymEnv


class CountRecall(POPGymEnv):
    """A game where the agent is queried on how many times it has observed
    an event in the past. This tests long-term order-agnostic memory like sets.

    Args:
        max_episode_length: The maximum number of timesteps in an episode
        error_clamp: Denotes the domain of the linear portion of the reward
            function. E.g. error_clamp == 2 means the agent will receive
            linearly-decreasing rewards for counts off by up to 2. Errors greater
            than 2 provide the same negative reward as an error of 2. Note as
            error_clamp -> inf, a crappy agent and perfect agent will acheive
            the same total reward.
        deck_type: What we use to count/differentiate cards.
            Can be colors, suits, or ranks.

    Returns:
        A gym environment
    """

    def __init__(self, num_decks=1, error_clamp=0.5, deck_type="colors"):
        self.value_deck = Deck(num_decks=num_decks)
        self.query_deck = Deck(num_decks=num_decks)
        self.value_deck.add_players("in_play")
        self.query_deck.add_players("in_play")
        if deck_type == "colors":
            self.deck_type = self.value_deck.colors
            self.deck_idx_type = self.value_deck.colors_idx
        elif deck_type == "suits":
            self.deck_type = self.value_deck.suits
            self.deck_idx_type = self.value_deck.suits_idx
        elif deck_type == "ranks":
            self.deck_type = self.value_deck.ranks
            self.deck_idx_type = self.value_deck.ranks_idx
        else:
            raise NotImplementedError(f"Invalid deck type {deck_type}")

        self.num_distinct_cards = len(np.unique(self.deck_type))
        self.max_card_count = self.value_deck.num_cards / self.num_distinct_cards
        # Space: [dealt card, card query]
        self.observation_space = gym.spaces.MultiDiscrete([self.num_distinct_cards] * 2)
        self.state_space = gym.spaces.Tuple(
            (
                gym.spaces.Box(0.0, 1.0, (self.num_distinct_cards,)),
                gym.spaces.Box(0.0, 1.0, (self.num_distinct_cards,)),
                # The current query is needed, but not the dealt card
                self.observation_space,
            )
        )
        self.last_obs = np.array([0, 0])

        self.action_space = gym.spaces.Discrete(self.max_card_count)
        self.max_episode_length = self.value_deck.num_cards - 1
        self.error_clamp = error_clamp
        self.reward_scale = 1 / self.max_episode_length

    def render(self):
        dealt = self.deck_type[self.value]
        queried = self.deck_type[self.query]
        print(f"Dealt {dealt}, recall {queried} count")

    def get_state(self):
        state = (
            (self.counts / self.max_card_count).astype(np.float32),
            (self.query_counts / self.max_card_count).astype(np.float32),
            self.last_obs.copy(),
        )
        return state

    def step(self, action):
        if isinstance(action, np.ndarray):
            action = action.item()

        self.prev_query = self.query
        prev_count = self.counts[self.prev_query]
        self.value, self.query = self.deal()
        self.counts[self.value] += 1
        self.query_counts[self.query] += 1

        reward = 1 if action == prev_count else -1
        reward *= self.reward_scale

        done = len(self.value_deck) == 0

        obs = np.array([self.value, self.query], dtype=np.int64)
        self.last_obs = obs
        info = {"counts": self.counts}

        return obs.copy(), reward, done, info

    def sample_deck(self):
        return np.random.choice(self.deck_idx_type, size=self.max_episode_length + 1)

    def deal(self):
        self.value_deck.deal("in_play")
        self.query_deck.deal("in_play")
        return (
            self.deck_idx_type[self.value_deck["in_play"][-1]],
            self.deck_idx_type[self.query_deck["in_play"][-1]],
        )

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, Dict[str, Any]]]:
        super().reset(seed=seed)
        self.value_deck.reset(rng=self.np_random)
        self.query_deck.reset(rng=self.np_random)  # Having another PRNG is not needed.

        self.counts = np.zeros((self.num_distinct_cards,), dtype=np.int64)
        self.query_counts = np.zeros((self.num_distinct_cards,), dtype=np.int64)

        self.value, self.query = self.deal()
        self.prev_query = None
        self.counts[self.value] += 1
        self.query_counts[self.query] += 1

        obs = np.array([self.value, self.query], dtype=np.int64)
        self.last_obs = obs.copy()
        if return_info:
            return obs, {"counts": self.counts}

        return obs


class CountRecallEasy(CountRecall):
    pass


class CountRecallMedium(CountRecall):
    def __init__(self, *args, **kwargs):
        super().__init__(num_decks=2, deck_type="suits")


class CountRecallHard(CountRecall):
    def __init__(self, *args, **kwargs):
        super().__init__(num_decks=4, deck_type="ranks")



from abc import abstractmethod

from gym import Env, spaces
from gym.core import ObsType

from popgym.core.observability import Observability


class POPGymEnv(Env):
    """A wrapper around gym.Env providing utilities for partial observability"""

    # The gym space for the underlying hidden Markov state
    state_space: spaces.Space[ObsType]
    # The observability level determines if and how the hidden Markov state
    # is returned from step() via the ObservabilityWrapper
    observability_level: Observability = Observability.PARTIAL
    # Whether the observation must contain the
    # previous action to learn the optimal policy.
    # If true, consider using the LastActionWrapper
    obs_requires_prev_action: bool = False

    @abstractmethod
    def get_state(self) -> ObsType:
        """Returns the underlying hidden Markov state"""


import copy

import gym
import numpy as np
import pandas as pd
import torch

from popgym.baselines.ray_models.ray_diffnc import DiffNC
from popgym.baselines.ray_models.ray_elman import Elman
from popgym.baselines.ray_models.ray_frameconv import Frameconv
from popgym.baselines.ray_models.ray_framestack import Framestack
from popgym.baselines.ray_models.ray_fwp import FastWeightProgrammer
from popgym.baselines.ray_models.ray_gru import GRU
from popgym.baselines.ray_models.ray_indrnn import IndRNN
from popgym.baselines.ray_models.ray_linear_attention import LinearAttention
from popgym.baselines.ray_models.ray_lmu import LMU
from popgym.baselines.ray_models.ray_lstm import LSTM
from popgym.baselines.ray_models.ray_mlp import MLP, BasicMLP
from popgym.baselines.ray_models.ray_s4d import S4D

BATCH = 64
TIME = 1024
DIM = 256
h = 128
SAMPLES = 10


def main():
    cfg = {
        "max_seq_len": TIME,
        "custom_model_config": {
            "hidden_size": DIM,
            "preprocessor_input_size": h,
            "preprocessor": torch.nn.Identity(),
            "preprocessor_output_size": h,
            "postprocessor": torch.nn.Identity(),
            "actor": torch.nn.Identity(),
            "critic": torch.nn.Identity(),
            "postprocessor_output_size": DIM,
        },
    }
    obs_shape = gym.spaces.Box(low=-np.inf, high=np.inf, shape=(128,), dtype=np.float32)
    act_shape = gym.spaces.Box(low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32)
    args = [obs_shape, act_shape, 1, cfg, "name"]
    rnn_cfg = copy.deepcopy(cfg)
    rnn_cfg["custom_model_config"]["benchmark"] = True
    rnn_args = [obs_shape, act_shape, 1, rnn_cfg, "name"]

    models = {
        "LSTM": [LSTM, rnn_args],
        "GRU": [GRU, rnn_args],
        "Elman": [Elman, rnn_args],
        "IndRNN": [IndRNN, args],
        "LMU": [LMU, args],
        "FART": [LinearAttention, args],
        "FWP": [FastWeightProgrammer, args],
        "S4D": [S4D, args],
        "MLP": [BasicMLP, args],
        "PosMLP": [MLP, args],
        "TCN": [Frameconv, args],
        "Fr.Stack": [Framestack, args],
        "DNC": [DiffNC, args],
    }
    results = []
    for name, (model, args) in models.items():
        print("TRAINING", name)
        results += train_closure(model(*args), "cuda")
        # print("# Params:", num_params)

    for name, (model, args) in models.items():
        print("EVALUATING", name)
        results += inference_closure(model(*args), "cpu")
        results += inference_closure(model(*args), "cuda")

    df = pd.DataFrame(results).sort_values(["mode", "device", "model"])
    df.to_csv("throughput.csv")
    # df.style.format(precision=2).hide_index().to_latex()
    breakpoint()


def train_closure(model, device):
    model = model.to(device)
    model.train()
    opt = torch.optim.Adam(model.parameters())
    data = [torch.rand((BATCH, TIME, h), device=device) for i in range(SAMPLES)]
    seq_lens = torch.full((BATCH,), TIME, device=device)
    state = model.get_initial_state()
    state = [s.unsqueeze(0).repeat(BATCH, *([1] * s.dim())).to(device) for s in state]

    # Warm up kernels
    for i in range(2):
        _, _ = model.forward(
            {"obs_flat": torch.rand_like(data[0].reshape(BATCH * TIME, -1))},
            state,
            seq_lens,
        )
    del _
    opt.zero_grad()
    torch.cuda.synchronize()
    base_mem = torch.cuda.memory_stats()["allocated_bytes.all.current"]
    num_params = sum([p.numel() for p in list(model.parameters()) if p.requires_grad])

    import time

    results = []
    torch.cuda.reset_peak_memory_stats()
    for i in range(SAMPLES):
        start = time.time()
        opt.zero_grad()
        out, _ = model.forward(
            {"obs_flat": data[i].reshape(BATCH * TIME, -1)}, state, seq_lens
        )
        loss = out.mean()
        loss.backward()
        opt.step()
        torch.cuda.synchronize()
        stop = time.time()
        results.append(
            {
                "model": model.__class__.__name__,
                "time": stop - start,
                "device": device,
                "mode": "train",
                "num_params": num_params,
            }
        )
    mem = torch.cuda.memory_stats()["allocated_bytes.all.peak"] - base_mem
    for r in results:
        r.update({"mem": mem})
    return results


def inference_closure(model, device):
    model = model.to(device)
    model.eval()
    data = [torch.rand((BATCH, 1, h), device=device) for i in range(SAMPLES)]
    seq_lens = torch.full((BATCH,), 1, device=device)
    state = model.get_initial_state()
    state = [s.unsqueeze(0).repeat(BATCH, *([1] * s.dim())).to(device) for s in state]

    # Warm up kernels
    with torch.no_grad():
        for i in range(2):
            _, _ = model.forward(
                {"obs_flat": torch.rand_like(data[0].reshape(BATCH, -1))},
                state,
                seq_lens,
            )
            del _
    if device != "cpu":
        torch.cuda.synchronize()

    import time

    num_params = sum([p.numel() for p in list(model.parameters()) if p.requires_grad])
    results = []
    torch.cuda.reset_peak_memory_stats()
    with torch.no_grad():
        for i in range(SAMPLES):
            start = time.time()
            for d in data:
                out, state = model.forward(
                    {"obs_flat": data[i].reshape(BATCH, -1)}, state, seq_lens
                )
                if device == "cuda":
                    torch.cuda.synchronize()
            stop = time.time()
            results.append(
                {
                    "model": model.__class__.__name__,
                    "time": (stop - start) / SAMPLES,
                    "device": device,
                    "mode": "inference",
                    "num_params": num_params,
                }
            )
    for r in results:
        r.update({"mem": 0})

    return results

    # print(
    #    f"{model.__class__.__name__} {device} sequence inference time:"
    #    f" {mu:.1f}+/-{std:.1f}ms memory: {mem / (1e6):.2f}MB"
    # )


if __name__ == "__main__":
    main()



import inspect
from typing import Any, Dict

import gym

from popgym.envs.autoencode import (
    Autoencode,
    AutoencodeEasy,
    AutoencodeHard,
    AutoencodeMedium,
)
from popgym.envs.battleship import (
    Battleship,
    BattleshipEasy,
    BattleshipHard,
    BattleshipMedium,
)
from popgym.envs.concentration import (
    Concentration,
    ConcentrationEasy,
    ConcentrationHard,
    ConcentrationMedium,
)
from popgym.envs.count_recall import (
    CountRecall,
    CountRecallEasy,
    CountRecallHard,
    CountRecallMedium,
)
from popgym.envs.higher_lower import (
    HigherLower,
    HigherLowerEasy,
    HigherLowerHard,
    HigherLowerMedium,
)
from popgym.envs.labyrinth_escape import (
    LabyrinthEscape,
    LabyrinthEscapeEasy,
    LabyrinthEscapeHard,
    LabyrinthEscapeMedium,
)
from popgym.envs.labyrinth_explore import (
    LabyrinthExplore,
    LabyrinthExploreEasy,
    LabyrinthExploreHard,
    LabyrinthExploreMedium,
)
from popgym.envs.minesweeper import (
    MineSweeper,
    MineSweeperEasy,
    MineSweeperHard,
    MineSweeperMedium,
)
from popgym.envs.multiarmed_bandit import (
    MultiarmedBandit,
    MultiarmedBanditEasy,
    MultiarmedBanditHard,
    MultiarmedBanditMedium,
)
from popgym.envs.noisy_stateless_cartpole import (
    NoisyStatelessCartPole,
    NoisyStatelessCartPoleEasy,
    NoisyStatelessCartPoleHard,
    NoisyStatelessCartPoleMedium,
)
from popgym.envs.noisy_stateless_pendulum import (
    NoisyStatelessPendulum,
    NoisyStatelessPendulumEasy,
    NoisyStatelessPendulumHard,
    NoisyStatelessPendulumMedium,
)
from popgym.envs.repeat_first import (
    RepeatFirst,
    RepeatFirstEasy,
    RepeatFirstHard,
    RepeatFirstMedium,
)
from popgym.envs.repeat_previous import (
    RepeatPrevious,
    RepeatPreviousEasy,
    RepeatPreviousHard,
    RepeatPreviousMedium,
)
from popgym.envs.stateless_cartpole import (
    StatelessCartPole,
    StatelessCartPoleEasy,
    StatelessCartPoleHard,
    StatelessCartPoleMedium,
)
from popgym.envs.stateless_pendulum import (
    StatelessPendulum,
    StatelessPendulumEasy,
    StatelessPendulumHard,
    StatelessPendulumMedium,
)

#
# Simple envs
#
SIMPLE_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    Autoencode: {"id": "popgym-Autoencode-v0"},
    RepeatPrevious: {"id": "popgym-RepeatPrevious-v0"},
    RepeatFirst: {"id": "popgym-RepeatFirst-v0"},
    CountRecall: {"id": "popgym-CountRecall-v0"},
}

SIMPLE_ENVS_EASY: Dict[gym.Env, Dict[str, Any]] = {
    AutoencodeEasy: {"id": "popgym-AutoencodeEasy-v0"},
    RepeatPreviousEasy: {"id": "popgym-RepeatPreviousEasy-v0"},
    RepeatFirstEasy: {"id": "popgym-RepeatFirstEasy-v0"},
    CountRecallEasy: {"id": "popgym-CountRecallEasy-v0"},
}

SIMPLE_ENVS_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    AutoencodeMedium: {"id": "popgym-AutoencodeMedium-v0"},
    RepeatPreviousMedium: {"id": "popgym-RepeatPreviousMedium-v0"},
    RepeatFirstMedium: {"id": "popgym-RepeatFirstMedium-v0"},
    CountRecallMedium: {"id": "popgym-CountRecallMedium-v0"},
}

SIMPLE_ENVS_HARD: Dict[gym.Env, Dict[str, Any]] = {
    AutoencodeHard: {"id": "popgym-AutoencodeHard-v0"},
    RepeatPreviousHard: {"id": "popgym-RepeatPreviousHard-v0"},
    RepeatFirstHard: {"id": "popgym-RepeatFirstHard-v0"},
    CountRecallHard: {"id": "popgym-CountRecallHard-v0"},
}

ALL_SIMPLE_ENVS = {**SIMPLE_ENVS_EASY, **SIMPLE_ENVS_MEDIUM, **SIMPLE_ENVS_HARD}

#
# Control envs
#
CONTROL_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    StatelessCartPole: {"id": "popgym-StatelessCartPole-v0"},
    StatelessPendulum: {
        "id": "popgym-StatelessPendulum-v0",
    },
    # BipedalWalker: {"id": "popgym-BipedalWalker-v0"},
}

CONTROL_ENVS_EASY: Dict[gym.Env, Dict[str, Any]] = {
    StatelessCartPoleEasy: {"id": "popgym-StatelessCartPoleEasy-v0"},
    StatelessPendulumEasy: {
        "id": "popgym-StatelessPendulumEasy-v0",
    },
    # BipedalWalkerEasy: {"id": "popgym-BipedalWalkerEasy-v0"},
}

CONTROL_ENVS_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    StatelessCartPoleMedium: {"id": "popgym-StatelessCartPoleMedium-v0"},
    StatelessPendulumMedium: {
        "id": "popgym-StatelessPendulumMedium-v0",
    },
    # BipedalWalkerMedium: {"id": "popgym-BipedalWalkerMedium-v0"},
}

CONTROL_ENVS_HARD: Dict[gym.Env, Dict[str, Any]] = {
    StatelessCartPoleHard: {"id": "popgym-StatelessCartPoleHard-v0"},
    StatelessPendulumHard: {
        "id": "popgym-StatelessPendulumHard-v0",
    },
    # BipedalWalkerHard: {"id": "popgym-BipedalWalkerHard-v0"},
}

ALL_CONTROL_ENVS = {**CONTROL_ENVS_EASY, **CONTROL_ENVS_MEDIUM, **CONTROL_ENVS_HARD}

#
# Noisy envs
#
NOISY_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    NoisyStatelessCartPole: {"id": "popgym-NoisyStatelessCartPole-v0"},
    NoisyStatelessPendulum: {
        "id": "popgym-NoisyStatelessPendulum-v0",
    },
    MultiarmedBandit: {"id": "popgym-MultiarmedBandit-v0"},
    # NonstationaryBandit: {"id": "popgym-NonstationaryBandit-v0"},
}

NOISY_ENVS_EASY: Dict[gym.Env, Dict[str, Any]] = {
    NoisyStatelessCartPoleEasy: {"id": "popgym-NoisyStatelessCartPoleEasy-v0"},
    NoisyStatelessPendulumEasy: {
        "id": "popgym-NoisyStatelessPendulumEasy-v0",
    },
    MultiarmedBanditEasy: {"id": "popgym-MultiarmedBanditEasy-v0"},
    # NonstationaryBanditEasy: {"id": "popgym-NonstationaryBanditEasy-v0"},
}

NOISY_ENVS_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    NoisyStatelessCartPoleMedium: {"id": "popgym-NoisyStatelessCartPoleMedium-v0"},
    NoisyStatelessPendulumMedium: {
        "id": "popgym-NoisyStatelessPendulumMedium-v0",
    },
    MultiarmedBanditMedium: {"id": "popgym-MultiarmedBanditMedium-v0"},
    # NonstationaryBanditMedium: {"id": "popgym-NonstationaryBanditMedium-v0"},
}

NOISY_ENVS_HARD: Dict[gym.Env, Dict[str, Any]] = {
    NoisyStatelessCartPoleHard: {"id": "popgym-NoisyStatelessCartPoleHard-v0"},
    NoisyStatelessPendulumHard: {
        "id": "popgym-NoisyStatelessPendulumHard-v0",
    },
    MultiarmedBanditHard: {"id": "popgym-MultiarmedBanditHard-v0"},
    # NonstationaryBanditHard: {"id": "popgym-NonstationaryBanditHard-v0"},
}

ALL_NOISY_ENVS = {**NOISY_ENVS_EASY, **NOISY_ENVS_MEDIUM, **NOISY_ENVS_HARD}

#
# Game envs
#
GAME_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    HigherLower: {"id": "popgym-HigherLower-v0"},
    Battleship: {"id": "popgym-Battleship-v0"},
    Concentration: {"id": "popgym-Concentration-v0"},
    MineSweeper: {"id": "popgym-MineSweeper-v0"},
}

GAME_ENVS_EASY: Dict[gym.Env, Dict[str, Any]] = {
    HigherLowerEasy: {"id": "popgym-HigherLowerEasy-v0"},
    BattleshipEasy: {"id": "popgym-BattleshipEasy-v0"},
    ConcentrationEasy: {"id": "popgym-ConcentrationEasy-v0"},
    MineSweeperEasy: {"id": "popgym-MineSweeperEasy-v0"},
}

GAME_ENVS_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    HigherLowerMedium: {"id": "popgym-HigherLowerMedium-v0"},
    BattleshipMedium: {"id": "popgym-BattleshipMedium-v0"},
    ConcentrationMedium: {"id": "popgym-ConcentrationMedium-v0"},
    MineSweeperMedium: {"id": "popgym-MineSweeperMedium-v0"},
}

GAME_ENVS_HARD: Dict[gym.Env, Dict[str, Any]] = {
    HigherLowerHard: {"id": "popgym-HigherLowerHard-v0"},
    BattleshipHard: {"id": "popgym-BattleshipHard-v0"},
    ConcentrationHard: {"id": "popgym-ConcentrationHard-v0"},
    MineSweeperHard: {"id": "popgym-MineSweeperHard-v0"},
}

ALL_GAME_ENVS = {**GAME_ENVS_EASY, **GAME_ENVS_MEDIUM, **GAME_ENVS_HARD}

#
# Navigation envs
#
NAVIGATION_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    LabyrinthExplore: {"id": "popgym-LabyrinthExplore-v0"},
    LabyrinthEscape: {"id": "popgym-LabyrinthEscape-v0"},
}

NAVIGATION_ENVS_EASY: Dict[gym.Env, Dict[str, Any]] = {
    LabyrinthExploreEasy: {"id": "popgym-LabyrinthExploreEasy-v0"},
    LabyrinthEscapeEasy: {"id": "popgym-LabyrinthEscapeEasy-v0"},
}

NAVIGATION_ENVS_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    LabyrinthExploreMedium: {"id": "popgym-LabyrinthExploreMedium-v0"},
    LabyrinthEscapeMedium: {"id": "popgym-LabyrinthEscapeMedium-v0"},
}

NAVIGATION_ENVS_HARD: Dict[gym.Env, Dict[str, Any]] = {
    LabyrinthExploreHard: {"id": "popgym-LabyrinthExploreHard-v0"},
    LabyrinthEscapeHard: {"id": "popgym-LabyrinthEscapeHard-v0"},
}

ALL_NAVIGATION_ENVS = {
    **NAVIGATION_ENVS_EASY,
    **NAVIGATION_ENVS_MEDIUM,
    **NAVIGATION_ENVS_HARD,
}

#
# Sets of envs
#
ALL_BASE_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    **SIMPLE_ENVS,
    **CONTROL_ENVS,
    **NOISY_ENVS,
    **GAME_ENVS,
    **NAVIGATION_ENVS,
}
ALL_EASY_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    **SIMPLE_ENVS_EASY,
    **CONTROL_ENVS_EASY,
    **NOISY_ENVS_EASY,
    **GAME_ENVS_EASY,
    **NAVIGATION_ENVS_EASY,
}
ALL_MEDIUM_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    **SIMPLE_ENVS_MEDIUM,
    **CONTROL_ENVS_MEDIUM,
    **NOISY_ENVS_MEDIUM,
    **GAME_ENVS_MEDIUM,
    **NAVIGATION_ENVS_MEDIUM,
}
ALL_HARD_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    **SIMPLE_ENVS_HARD,
    **CONTROL_ENVS_HARD,
    **NOISY_ENVS_HARD,
    **GAME_ENVS_HARD,
    **NAVIGATION_ENVS_HARD,
}
ALL_ENVS: Dict[gym.Env, Dict[str, Any]] = {
    **ALL_EASY_ENVS,
    **ALL_MEDIUM_ENVS,
    **ALL_HARD_ENVS,
}

# Register envs
for e, v in ALL_ENVS.items():
    mod_name = inspect.getmodule(e).__name__  # type: ignore
    gym.envs.register(entry_point=":".join([mod_name, e.__name__]), order_enforce=False, **v)



import enum

OBS = "obs"
STATE = "state"


class Observability(enum.IntEnum):
    """Defines the observability level of the environment.

    To be used with popgym.wrappers.ObservabilityWrapper.

    - PARTIAL: Partial observation, the Markov state is not returned.
    - FULL_IN_INFO_DICT: Partial observation, the Markov state is provided
        by the info dict.
    - FULL: The Markov state is used as observation.
    - FULL_AND_PARTIAL: The observation is a dict {"obs": obs, "state": state}.
    """

    PARTIAL = 0
    FULL_IN_INFO_DICT = 1
    FULL = 2
    FULL_AND_PARTIAL = 3



import pytest
from gym.utils.env_checker import check_env

from popgym import ALL_ENVS
from popgym.core.observability import OBS, STATE, Observability
from popgym.wrappers.antialias import Antialias
from popgym.wrappers.markovian import Markovian
from popgym.wrappers.previous_action import PreviousAction


def check_space(space, data):
    valid = space.contains(data)
    if not valid:
        raise ValueError(f"space {space} does not contain data {data}")


@pytest.mark.parametrize("env", ALL_ENVS.keys())
def test_previousaction_step(env):
    wrapped_noaa = PreviousAction(env())
    wrapped_noaa.reset()
    check_env(wrapped_noaa)


@pytest.mark.parametrize("env", ALL_ENVS.keys())
def test_antialias_step(env):
    wrapped_aa = Antialias(env())
    wrapped_aa.reset()
    check_env(wrapped_aa)


@pytest.mark.parametrize("env", ALL_ENVS.keys())
def test_previousaction_antialias_step(env):
    wrapped_aa = Antialias(PreviousAction(env()))
    wrapped_aa.reset()
    check_env(wrapped_aa)


@pytest.mark.parametrize("env", ALL_ENVS.keys())
def test_markovian_state_space_full(env):
    wrapped = Markovian(env(), Observability.FULL)
    obs = wrapped.reset()
    check_space(wrapped.observation_space, obs)
    check_space(wrapped.state_space, obs)
    for i in range(10):
        obs, reward, done, info = wrapped.step(wrapped.action_space.sample())
        check_space(wrapped.observation_space, obs)
        check_space(wrapped.state_space, obs)
        if done:
            _ = wrapped.reset()


@pytest.mark.parametrize("env", ALL_ENVS.keys())
def test_markovian_state_space_partial(env):
    e = env()
    wrapped = Markovian(e, Observability.PARTIAL)
    obs = wrapped.reset()
    check_space(wrapped.observation_space, obs)
    check_space(e.observation_space, obs)
    for i in range(10):
        obs, reward, done, info = wrapped.step(wrapped.action_space.sample())
        check_space(wrapped.observation_space, obs)
        check_space(e.observation_space, obs)
        if done:
            _ = wrapped.reset()


@pytest.mark.parametrize("env", ALL_ENVS.keys())
def test_markovian_state_space_info_dict(env):
    e = env()
    wrapped = Markovian(e, Observability.FULL_IN_INFO_DICT)
    wrapped.reset()
    for i in range(10):
        obs, reward, done, info = wrapped.step(wrapped.action_space.sample())
        check_space(wrapped.state_space, info[STATE])
        if done:
            _ = wrapped.reset()


@pytest.mark.parametrize("env", ALL_ENVS.keys())
def test_state_space_full_and_partial(env):
    e = env()
    wrapped = Markovian(e, Observability.FULL_AND_PARTIAL)
    obs = wrapped.reset()
    check_space(wrapped.observation_space[STATE], obs[STATE])
    check_space(wrapped.observation_space[OBS], obs[OBS])
    check_space(e.observation_space, obs[OBS])
    for i in range(10):
        obs, reward, done, info = wrapped.step(wrapped.action_space.sample())
        check_space(wrapped.observation_space[STATE], obs[STATE])
        check_space(wrapped.observation_space[OBS], obs[OBS])
        check_space(e.observation_space, obs[OBS])



from typing import Optional, Tuple, Union

import gym
import numpy as np

from popgym.envs.stateless_pendulum import StatelessPendulum


class NoisyStatelessPendulum(StatelessPendulum):
    def __init__(self, *args, **kwargs):
        noise_sigma = kwargs.pop("noise_sigma", 0.1)
        super().__init__(*args, **kwargs)
        self.noise_sigma = np.full(self.observation_space.shape, noise_sigma)

    def add_noise_to_obs(self, obs):
        noise = self.np_random.normal(0, self.noise_sigma).astype(np.float32)
        obs_space = self.observation_space
        obs = np.clip(
            obs + noise, obs_space.low, obs_space.high
        )
        return obs

    def step(
        self, action: gym.core.ActType
    ) -> Tuple[gym.core.ObsType, float, bool, dict]:
        next_obs, reward, done, info = super().step(action)
        return self.add_noise_to_obs(next_obs), reward, done, info

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, dict]]:

        if return_info:
            init_obs, info = super().reset(
                seed=seed, return_info=return_info, options=options
            )
            init_obs = self.add_noise_to_obs(init_obs)
            return init_obs, info
        else:
            init_obs = super().reset(
                seed=seed, return_info=return_info, options=options
            )
            init_obs = self.add_noise_to_obs(init_obs)
            return init_obs


class NoisyStatelessPendulumEasy(NoisyStatelessPendulum):
    def __init__(self):
        super().__init__(noise_sigma=0.1)


class NoisyStatelessPendulumMedium(NoisyStatelessPendulum):
    def __init__(self):
        super().__init__(noise_sigma=0.2)


class NoisyStatelessPendulumHard(NoisyStatelessPendulum):
    def __init__(self):
        super().__init__(noise_sigma=0.3)



import unittest

import numpy as np

from popgym.envs.minesweeper import HiddenSquare, MineSweeper
from tests.base_env_test import AbstractTest


class TestMineSweeper(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = MineSweeper()

    def test_perfect(self):
        self.env.reset()
        xs, ys = np.where(self.env.hidden_grid == HiddenSquare.CLEAR)
        cum_rew = 0
        done = False
        for x, y in zip(xs, ys):
            self.assertFalse(done)
            obs, reward, done, info = self.env.step(np.array([x, y]))
            self.assertAlmostEqual(reward, self.env.success_reward_scale)
            cum_rew += reward

        self.assertTrue(done)
        self.assertAlmostEqual(cum_rew, 1.0)

    def test_worst(self):
        self.env.reset()
        ts = self.env.max_episode_length - 1
        xs, ys = np.where(self.env.hidden_grid == HiddenSquare.CLEAR)
        cum_rew = 0
        done = False
        for i in range(ts):
            action = np.array([xs[0], ys[0]])
            obs, reward, done, info = self.env.step(action)
            cum_rew += reward

        self.assertFalse(done)

        xs, ys = np.where(self.env.hidden_grid == HiddenSquare.MINE)
        action = np.array([xs[0], ys[0]])
        obs, reward, done, info = self.env.step(action)
        cum_rew += reward
        self.assertTrue(done)

        self.assertAlmostEqual(cum_rew, -1.0)

        cum_rew = 0
        for x, y in zip(xs, ys):
            obs, reward, done, info = self.env.step(np.array([x, y]))
            cum_rew += reward



# Based on https://github.com/hrshtv/pytorch-lmu

import numpy as np
import torch
from scipy.signal import cont2discrete
from torch import fft, nn
from torch.nn import functional as F
from torch.nn import init


def leCunUniform(tensor):
    """
    LeCun Uniform Initializer
    """
    fan_in = init._calculate_correct_fan(tensor, "fan_in")
    limit = np.sqrt(3.0 / fan_in)
    init.uniform_(
        tensor, -limit, limit
    )  # fills the tensor with values sampled from U(-limit, limit)


class LMUCell(nn.Module):
    """LMU Cell.

    Parameters:
        input_size (int) :
            Size of the input vector (x_t)
        hidden_size (int) :
            Size of the hidden vector (h_t)
        memory_size (int) :
            Size of the memory vector (m_t)
        theta (int) :
            The number of timesteps in the sliding window that is
            represented using the LTI system
        learn_a (boolean) :
            Whether to learn the matrix A (default = False)
        learn_b (boolean) :
            Whether to learn the matrix B (default = False)
    """

    def __init__(
        self, input_size, hidden_size, memory_size, theta, learn_a=False, learn_b=False
    ):
        super(LMUCell, self).__init__()

        self.hidden_size = hidden_size
        self.memory_size = memory_size
        self.f = nn.Tanh()

        A, B = self.stateSpaceMatrices(memory_size, theta)
        A = torch.from_numpy(A).float()
        B = torch.from_numpy(B).float()

        if learn_a:
            self.A = nn.Parameter(A)
        else:
            self.register_buffer("A", A)

        if learn_b:
            self.B = nn.Parameter(B)
        else:
            self.register_buffer("B", B)

        # Declare Model parameters:
        # Encoding vectors
        self.e_x = nn.Parameter(torch.empty(1, input_size))
        self.e_h = nn.Parameter(torch.empty(1, hidden_size))
        self.e_m = nn.Parameter(torch.empty(1, memory_size))
        # Kernels
        self.W_x = nn.Parameter(torch.empty(hidden_size, input_size))
        self.W_h = nn.Parameter(torch.empty(hidden_size, hidden_size))
        self.W_m = nn.Parameter(torch.empty(hidden_size, memory_size))

        self.initParameters()

    def initParameters(self):
        """Initialize the cell's parameters."""

        # Initialize encoders
        leCunUniform(self.e_x)
        leCunUniform(self.e_h)
        init.constant_(self.e_m, 0)
        # Initialize kernels
        init.xavier_normal_(self.W_x)
        init.xavier_normal_(self.W_h)
        init.xavier_normal_(self.W_m)

    def stateSpaceMatrices(self, memory_size, theta):
        """Returns the discretized state space matrices A and B."""

        Q = np.arange(memory_size, dtype=np.float64).reshape(-1, 1)
        R = (2 * Q + 1) / theta
        i, j = np.meshgrid(Q, Q, indexing="ij")

        # Continuous
        A = R * np.where(i < j, -1, (-1.0) ** (i - j + 1))
        B = R * ((-1.0) ** Q)
        C = np.ones((1, memory_size))
        D = np.zeros((1,))

        # Convert to discrete
        A, B, C, D, dt = cont2discrete(system=(A, B, C, D), dt=1.0, method="zoh")

        return A, B

    def forward(self, x, state):
        """
        Parameters:
            x (torch.tensor):
                Input of size [batch_size, input_size]
            state (tuple):
                h (torch.tensor) : [batch_size, hidden_size]
                m (torch.tensor) : [batch_size, memory_size]
        """

        h, m = state

        # Equation (7) of the paper
        u = (
            F.linear(x, self.e_x) + F.linear(h, self.e_h) + F.linear(m, self.e_m)
        )  # [batch_size, 1]

        # Equation (4) of the paper
        m = F.linear(m, self.A) + F.linear(u, self.B)  # [batch_size, memory_size]

        # Equation (6) of the paper
        h = self.f(
            F.linear(x, self.W_x) + F.linear(h, self.W_h) + F.linear(m, self.W_m)
        )  # [batch_size, hidden_size]

        return h, m


class LMU(nn.Module):
    """LMU layer.

    Parameters:
        input_size (int) :
            Size of the input vector (x_t)
        hidden_size (int) :
            Size of the hidden vector (h_t)
        memory_size (int) :
            Size of the memory vector (m_t)
        theta (int) :
            The number of timesteps in the sliding window
            that is represented using the LTI system
        learn_a (boolean) :
            Whether to learn the matrix A (default = False)
        learn_b (boolean) :
            Whether to learn the matrix B (default = False)
    """

    def __init__(
        self, input_size, hidden_size, memory_size, theta, learn_a=False, learn_b=False
    ):
        super(LMU, self).__init__()
        self.hidden_size = hidden_size
        self.memory_size = memory_size
        self.cell = LMUCell(
            input_size, hidden_size, memory_size, theta, learn_a, learn_b
        )

    def forward(self, x, state=None):
        """
        Parameters:
            x (torch.tensor):
                Input of size [batch_size, seq_len, input_size]
            state (tuple) : (default = None)
                h (torch.tensor) : [batch_size, hidden_size]
                m (torch.tensor) : [batch_size, memory_size]
        """

        # Assuming batch dimension is always first,
        # followed by seq. length as the second dimension
        batch_size = x.size(0)
        seq_len = x.size(1)

        # Initial state (h_0, m_0)
        if state is None:
            h_0 = torch.zeros(batch_size, self.hidden_size)
            m_0 = torch.zeros(batch_size, self.memory_size)
            if x.is_cuda:
                h_0 = h_0.cuda()
                m_0 = m_0.cuda()
            state = (h_0, m_0)

        # Iterate over the timesteps
        output = []
        for t in range(seq_len):
            x_t = x[:, t, :]  # [batch_size, input_size]
            h_t, m_t = self.cell(x_t, state)
            state = (h_t, m_t)
            output.append(h_t)

        output = torch.stack(output)  # [seq_len, batch_size, hidden_size]
        output = output.permute(1, 0, 2)  # [batch_size, seq_len, hidden_size]

        return output, state  # state is (h_n, m_n) where n = seq_len


class LMUFFT(nn.Module):
    """Parallelized LMU Layer.

    Parameters:
        input_size (int) :
            Size of the input vector (x_t)
        hidden_size (int) :
            Size of the hidden vector (h_t)
        memory_size (int) :
            Size of the memory vector (m_t)
        seq_len (int) :
            Size of the sequence length (n)
        theta (int) :
            The number of timesteps in the sliding
            window that is represented using the LTI system
    """

    def __init__(self, input_size, hidden_size, memory_size, seq_len, theta):
        super(LMUFFT, self).__init__()

        self.hidden_size = hidden_size
        self.memory_size = memory_size
        self.seq_len = seq_len
        self.theta = theta

        self.W_u = nn.Linear(in_features=input_size, out_features=1)
        self.f_u = nn.ReLU()
        self.W_h = nn.Linear(
            in_features=memory_size + input_size, out_features=hidden_size
        )
        self.f_h = nn.ReLU()

        A, B = self.stateSpaceMatrices()
        self.register_buffer("A", A)  # [memory_size, memory_size]
        self.register_buffer("B", B)  # [memory_size, 1]

        H, fft_H = self.impulse()
        self.register_buffer("H", H)  # [memory_size, seq_len]
        self.register_buffer("fft_H", fft_H)  # [memory_size, seq_len + 1]

    def stateSpaceMatrices(self):
        """Returns the discretized state space matrices A and B."""

        Q = np.arange(self.memory_size, dtype=np.float64).reshape(-1, 1)
        R = (2 * Q + 1) / self.theta
        i, j = np.meshgrid(Q, Q, indexing="ij")

        # Continuous
        A = R * np.where(i < j, -1, (-1.0) ** (i - j + 1))
        B = R * ((-1.0) ** Q)
        C = np.ones((1, self.memory_size))
        D = np.zeros((1,))

        # Convert to discrete
        A, B, C, D, dt = cont2discrete(system=(A, B, C, D), dt=1.0, method="zoh")

        # To torch.tensor
        A = torch.from_numpy(A).float()  # [memory_size, memory_size]
        B = torch.from_numpy(B).float()  # [memory_size, 1]

        return A, B

    def impulse(self):
        """Returns the matrices H and the 1D Fourier transform of H (Equations
        23, 26 of the paper)"""

        H = []
        A_i = torch.eye(self.memory_size)
        for t in range(self.seq_len):
            H.append(A_i @ self.B)
            A_i = self.A @ A_i

        H = torch.cat(H, dim=-1)  # [memory_size, seq_len]
        fft_H = fft.rfft(H, n=2 * self.seq_len, dim=-1)  # [memory_size, seq_len + 1]

        return H, fft_H

    def forward(self, x):
        """
        Parameters:
            x (torch.tensor):
                Input of size [batch_size, seq_len, input_size]
        """

        batch_size, seq_len, input_size = x.shape

        # Equation 18 of the paper
        u = self.f_u(self.W_u(x))  # [batch_size, seq_len, 1]

        # Equation 26 of the paper
        fft_input = u.permute(0, 2, 1)  # [batch_size, 1, seq_len]
        fft_u = fft.rfft(
            fft_input, n=2 * seq_len, dim=-1
        )  # [batch_size, seq_len, seq_len+1]

        # Element-wise multiplication (uses broadcasting)
        # [batch_size, 1, seq_len+1] * [1, memory_size, seq_len+1]
        temp = fft_u * self.fft_H.unsqueeze(0)  # [batch_size, memory_size, seq_len+1]

        m = fft.irfft(
            temp, n=2 * seq_len, dim=-1
        )  # [batch_size, memory_size, seq_len+1]
        m = m[:, :, :seq_len]  # [batch_size, memory_size, seq_len]
        m = m.permute(0, 2, 1)  # [batch_size, seq_len, memory_size]

        # Equation 20 of the paper (W_m@m + W_x@x  W@[m;x])
        input_h = torch.cat(
            (m, x), dim=-1
        )  # [batch_size, seq_len, memory_size + input_size]
        h = self.f_h(self.W_h(input_h))  # [batch_size, seq_len, hidden_size]

        h_n = h[:, -1, :]  # [batch_size, hidden_size]

        return h, h_n



# Inspired by ray rllib at
# https://github.com/ray-project/ray/blob/master/rllib/examples/env/stateless_pendulum.py

from typing import Optional, Tuple, Union

import gym
import numpy as np
from gym.envs.classic_control import PendulumEnv
from gym.spaces import Box

from popgym.core.env import POPGymEnv


class StatelessPendulum(PendulumEnv, POPGymEnv):
    """Partially observable variant of the Pendulum gym environment.
    https://github.com/openai/gym/blob/master/gym/envs/classic_control/
    pendulum.py
    We delete the angular velocity component of the state, so that it
    can only be solved by a memory enhanced model (policy).

    Args:
        max_episode_length: Exactly what it sounds like
    Returns:
        A gym environment
    """

    def __init__(self, *args, **kwargs):
        self.max_episode_length = kwargs.pop("max_episode_length", 200)
        super().__init__(*args, **kwargs)

        # Fix our observation-space (remove angular velocity component).
        high = np.array([1.0, 1.0], dtype=np.float32)
        self.state_space = self.observation_space
        self.observation_space = Box(low=-high, high=high, dtype=np.float32)

    def reward_transform(self, reward):
        low, high = -16.2736044, 0
        shifted = reward + (high - low) / 2
        scaled = shifted / ((high - low) / 2)
        transformed = scaled / self.max_episode_length
        return transformed

    def get_state(self):
        return self._state

    def step(
        self, action: gym.core.ActType
    ) -> Tuple[gym.core.ObsType, float, bool, dict]:
        next_obs, reward, done, info = super().step(action)
        self._state = next_obs
        self.num_steps += 1
        if self.num_steps >= self.max_episode_length:
            done = True
        reward = self.reward_transform(reward)
        # next_obs is [cos(theta), sin(theta), theta-dot (angular velocity)]
        return next_obs[:-1], reward, done, info

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, dict]]:
        self.num_steps = 0
        if return_info:
            init_obs, info = super().reset(
                seed=seed, return_info=return_info, options=options
            )
            self._state = init_obs
            return init_obs[:-1], info

        else:
            init_obs = super().reset(
                seed=seed, return_info=return_info, options=options
            )
            self._state = init_obs
            # init_obs is [cos(theta), sin(theta), theta-dot (angular velocity)]
            return init_obs[:-1]


class StatelessPendulumEasy(StatelessPendulum):
    pass


class StatelessPendulumMedium(StatelessPendulum):
    def __init__(self):
        super().__init__(max_episode_length=150)


class StatelessPendulumHard(StatelessPendulum):
    def __init__(self):
        super().__init__(max_episode_length=100)



import unittest

from popgym.envs.repeat_previous import RepeatPrevious
from tests.base_env_test import AbstractTest


class TestRepeatPrevious(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = RepeatPrevious()

    def test_k(self):
        e = RepeatPrevious(k=2)
        obs0 = e.reset()
        obs1, rew1, done1, info1 = e.step(0)
        self.assertFalse(done1)
        obs2, rew2, done2, info2 = e.step(obs0)
        self.assertFalse(done2)
        obs3, rew3, done3, info3 = e.step(obs1)
        self.assertFalse(done3)

    def test_full(self):
        e = RepeatPrevious(k=2)
        obs1 = e.reset()
        cum_rew = 0
        obs2, reward, done, info = e.step(0)
        cum_rew += reward

        all_obs = [obs1, obs2]

        iters = 2
        while not done:
            a = all_obs[-2]
            obs, reward, done, info = e.step(a)
            all_obs.append(obs)
            cum_rew += reward
            iters += 1

        self.assertAlmostEqual(cum_rew, 1.0)

        self.assertEqual(iters, 52)

    def test_full_16(self):
        e = RepeatPrevious(k=16)
        all_obs = [e.reset()]
        iters = 1
        cum_rew = 0
        while iters < 16:
            obs, reward, done, info = e.step(0)
            self.assertFalse(done)
            self.assertEqual(reward, 0)
            cum_rew += reward
            all_obs.append(obs)
            iters += 1

        while iters < 52:
            a = all_obs[-16]
            self.assertFalse(done)
            obs, reward, done, info = e.step(a)
            cum_rew += reward
            all_obs.append(obs)
            iters += 1

        self.assertTrue(done)
        self.assertAlmostEqual(cum_rew, 1.0)
